//=============================================================================
//  Unified IO - C++ interface for industrial IO cards
//  Copyright (C) 2000, UAB "BBD SOFT" ( http://www.bbdsoft.com/ )
//
// This material is provided "as is", with absolutely no warranty expressed
// or implied. Any use is at your own risk.
//
// Permission to use or copy this software for any purpose is hereby granted
// without fee, provided the above notices are retained on all copies.
// Permission to modify the code and to distribute modified code is granted,
// provided the above notices are retained, and a notice that the code was
// modified is included with the above copyright notice.
//
//  The author of this program may be contacted at developers@bbdsoft.com
//=============================================================================

#ifndef _BTRACE_
   #include "btrace.hpp"
#endif

#ifndef _SST5136PFBPCI_
   #include "sst5136pfbpci.hpp"
#endif

#ifndef _STDEXCEPT_
   #include <stdexcept>
#endif

#ifndef _IOTRACK_
   #include "iotrack.hpp"
#endif

#ifndef _PCIBUS_
   #include "pcibus.hpp"
#endif

#ifndef _OSIFACE_
   #include "osiface.hpp"
#endif

#ifndef _FWFBUFF_
   #include "fwfbuff.hpp"
#endif

using namespace std;
using namespace Uniio;

//-----------------------------------------------------------------------------
SST5136PFBPCI::SST5136PFBPCI ( const char * const inFirmwarePath
                                     , const unsigned long inBaudRate
                                     , const unsigned inCardId
                                     )
  : IOCard ("SST-5136-PFB-PCI")
  , theMemory (NULL)
  , theBaudRate (inBaudRate)
{
BFUNCTRACE_DEVELOP();

const PCIDeviceInfo * anInfo = PCIBus::pciBus().getDeviceInfo (0x133D, 0x1000, inCardId );
if (!anInfo)
{
   throw runtime_error ("SST-5136-PFB-PCI: Card not found in the system.");
} // endif

theBaseAddress = anInfo -> baseAddress(0) & (~0x0f);
theIrq =  anInfo -> interruptLine();
theMemory = new PhysicalMemory (theBaseAddress, pfbMemorySize);

   if ((theMemory -> readChar (0x1FFFFF) & 0xFF) == 0xBA)
   {

      FirmwareFileBuffer aFirmwareFile
          ( string(string(inFirmwarePath) + string ("pciprofi.ss1")).data());
      theMemory -> writeMemoryBlock ( aFirmwareFile
                                    , aFirmwareFile.size()
                                    , 0x1000
                                    );

      theMemory -> writeChar (0x1FFFFF, 0x01);

      while ( theMemory -> readChar (0x1FFFFF) != 0x5A)
      {
         OSInterface::sleep (1000);
      }

      theMemory -> writeChar (0x1FFFFF, 0x00);

      while ( state() != cmdOFF)
      {
         OSInterface::sleep (20000);
      }

   } // end if state == 0   
   if (  ((theMemory -> readShort(pfbCardId) & 0xFFFF) != 0xAAD0)
      || ((theMemory -> readShort(pfbModId) & 0xFFFF) != 0xBB01)
      )
   {
      runtime_error exc ("SST-5136-PFB-PCI: wrong card id or wrong module id.");
      throw exc;
   }

   if (state() == cmdON)
   {
      executeCommand (cmdGO_OFF);
   }
   executeCommand (cmdREINIT);

   theInputsOffset = pageOffset (theMemory -> readChar(pfbMasRxPage));
   theOutputsOffset = pageOffset (theMemory -> readChar(pfbMasTxPage));
   theMasterControlOffset = pageOffset (theMemory -> readChar(pfbMasCntrlPage));

} // end SST5136PFBPCI constructor


//-----------------------------------------------------------------------------
SST5136PFBPCI::~SST5136PFBPCI()
{
BFUNCTRACE_DEVELOP();
delete theMemory;
} // end SST5136PFBPCI destructor

//-----------------------------------------------------------------------------
SST5136PFBPCI & SST5136PFBPCI::setBaudRate(const unsigned char inBaudRate)
{
BFUNCTRACE_DEVELOP();
BASSERT (inBaudRate <= baud12m );

   theBaudRate = inBaudRate;
   theMemory -> writeChar ( pfbBaud, theBaudRate);
return *this;
} // end SST5136PFBPCI 

//-----------------------------------------------------------------------------
SST5136PFBPCI & SST5136PFBPCI::autoconfigure()
{
BFUNCTRACE_DEVELOP();
   if (state() == cmdON)
   {
      executeCommand (cmdGO_OFF);
   }
      unsigned short modVersion = theMemory -> readShort(pfbModVer);
      if (modVersion < 0x0155)
      {
         runtime_error exc (asString("SST-5136-PFB-PCI: your pciprofi.ss1 module version is %X. You need version higher than 0155 to use autoconfigure feature", modVersion));
         throw exc;
      }
      executeCommand (cmdCLR_CFG_BUF);
      theMemory -> writeChar ( pfbStnAddr, 0);
      theMemory -> writeChar ( pfbActive, 1);
      theMemory -> writeChar ( pfbBaud, theBaudRate);
      theMemory -> writeChar ( pfbMasCntrlCfg, pfbMAS_CTRL_RUN_MODE
                                    | pfbMAS_CTRL_ENABLE);

      executeCommand (cmdAUTODP_CFG);
return *this;
} // end SST5136PFBPCI 

//-----------------------------------------------------------------------------
SST5136PFBPCI & SST5136PFBPCI::configurationToFlash()
{
BFUNCTRACE_DEVELOP();
   if (state() == cmdON)
   {
      executeCommand (cmdGO_OFF);
   }
   executeCommand (cmdPGM_TO_FLASH);
return *this;
} // end SST5136PFBPCI 

//-----------------------------------------------------------------------------
SST5136PFBPCI & SST5136PFBPCI::configurationFromFlash()
{
BFUNCTRACE_DEVELOP();
   if (state() == cmdON)
   {
      executeCommand (cmdGO_OFF);
   }
   executeCommand (cmdCFG_FROM_FLASH);
return *this;
} // end SST5136PFBPCI 

//-----------------------------------------------------------------------------
SST5136PFBPCI & SST5136PFBPCI::online()
{
BFUNCTRACE_DEVELOP();
   theMemory -> writeChar ( pfbStnAddr, 0);
   theMemory -> writeChar ( pfbActive, 1);
   theMemory -> writeChar ( pfbBaud, theBaudRate);
   theMemory -> writeChar ( pfbMasCntrlCfg, pfbMAS_CTRL_RUN_MODE
                                 | pfbMAS_CTRL_ENABLE);
   executeCommand (cmdGO_ON);
return *this;
} // end SST5136PFBPCI 

//-----------------------------------------------------------------------------
SST5136PFBPCI & SST5136PFBPCI::configure(const char * const inConfigFile)
{
BFUNCTRACE_DEVELOP();
   if (state() == cmdON)
   {
      executeCommand (cmdGO_OFF);
   }
   FirmwareFileBuffer aFirmwareFile ( inConfigFile );

   unsigned long anOffset = pageOffset (theMemory -> readChar(pfbBinCfgPage));

   unsigned index = 0;
   unsigned bytesLeft = aFirmwareFile.size();
   unsigned char aBlockNumber = 0;

   while (bytesLeft > 0)
   {
      if (bytesLeft >= maxPageSize)
      {
         theMemory -> writeMemoryBlock ( aFirmwareFile+index
                                       , maxPageSize
                                       , anOffset
                                       );
         theMemory -> writeChar ( pfbBinCfgLen, maxPageSize );

         bytesLeft -= maxPageSize;
         index += maxPageSize;
      }
      else
      {
         theMemory -> writeMemoryBlock ( aFirmwareFile+index
                                       , maxPageSize
                                       , anOffset
                                       );
         theMemory -> writeChar ( pfbBinCfgLen, bytesLeft );

         index += bytesLeft;
         bytesLeft = 0;
      } // end if
      theMemory -> writeChar ( pfbBinCfgOfs, aBlockNumber );
      executeCommand(cmdCPY_MAS_CFG);
      aBlockNumber++;
   } // end while

   executeCommand(cmdCFG_2BF_SHRAM);

return *this;
} // end SST5136PFBPCI 

//-----------------------------------------------------------------------------
IOCard & SST5136PFBPCI::updateInputs ( const IOTrack & inTrack
                                   , long * inImage
                                   , long & inTrackState
                                   )
{
BFUNCTRACE_DEVELOP ();

unsigned long aSlaveId = (inTrack.startingPoint() >> 16) & 0xFFFF;
unsigned short aCrossReff = theMemory -> readShort ( pfbMasStsTab
                                                  + (aSlaveId*2)
                                                  );

unsigned long aMasterControlOffset = theMasterControlOffset
                                   + ((aCrossReff & 0xFF) * masSizeOfTable);

   if (aCrossReff & 0x8000)
   {
      inTrackState = IOTrack::noFault;
   }
   else
   {
      inTrackState = IOTrack::moduleFault; 
   }

if (inTrack.hasPointOfType(IOTrack::digitalInput))
{

unsigned long aModuleId = inTrack.startingPoint() & 0xFFFF;

unsigned long anInputsOffset = theInputsOffset;

anInputsOffset += theMemory -> readShort ( aMasterControlOffset
                                         + masRxDataOfs
                                         );
anInputsOffset += aModuleId;

for ( int i = 0
    ; i < inTrack.numberOfPoints()/pointsPerBank
    ; i++)
{
   unsigned char newValue = theMemory -> readChar ( anInputsOffset
                                                  + i
                                                  );

   int imageBase = i*pointsPerBank;
   for (int j = 0; j<pointsPerBank; j++)
   {
      if ( inTrack.pointType(imageBase+j) == IOTrack::digitalInput )
      {
         inImage[imageBase+j] = newValue %2;
      }
      newValue /= 2;
   } // endfor
} // endfor
}
return * this;
} // end SST5136PFBPCI::updateInputs (..)


//-----------------------------------------------------------------------------
IOCard & SST5136PFBPCI::updateOutputs ( const IOTrack & inTrack
                                    , const long * const inImage
                                    , long & inTrackState
                                    )
{
BFUNCTRACE_DEVELOP ();
unsigned long aSlaveId = (inTrack.startingPoint() >> 16) & 0xFFFF;
unsigned short aCrossReff = theMemory -> readShort ( pfbMasStsTab
                                                  + (aSlaveId*2)
                                                  );

unsigned long aMasterControlOffset = theMasterControlOffset
                                   + ((aCrossReff & 0xFF) * masSizeOfTable);

   if (aCrossReff & 0x8000)
   {
      inTrackState = IOTrack::noFault;
   }
   else
   {
      inTrackState = IOTrack::moduleFault; 
   }

if (inTrack.hasPointOfType(IOTrack::digitalOutput))
{

unsigned long aModuleId = inTrack.startingPoint() & 0xFFFF;

unsigned long anOutputsOffset = theOutputsOffset;

anOutputsOffset += theMemory -> readShort ( aMasterControlOffset
                                         + masTxDataOfs
                                         );
anOutputsOffset += aModuleId;

   for ( int i = 0
       ; i < inTrack.numberOfPoints()/pointsPerBank
       ; i++)
   {
      unsigned char newValue = 0;

      int imageBase = i*pointsPerBank;

      for (int j = 0; j<pointsPerBank; j++)
      {
         if (  (inTrack.pointType(imageBase+j) == IOTrack::digitalOutput)
            && inImage[imageBase+j] )
         {
            newValue |= (1 << j);
         } // endif
      } // endfor
      theMemory -> writeChar ( anOutputsOffset + i, newValue);
   } // endfor
}
return * this;
} // end SST5136PFBPCI::updateOutputs (..)


//-----------------------------------------------------------------------------
IOCard & SST5136PFBPCI::addTrack ( IOTrack & inTrack )
{
BFUNCTRACE_DEVELOP ();

if (   inTrack.hasPointOfType ('U')
    || inTrack.hasPointOfType ('A') 
    || inTrack.hasPointOfType ('C') 
    || inTrack.hasPointOfType ('R') 
   )
{
   throw logic_error ("SST-5136-PFB-PCI: Current implementation does not support analog I/O and counters.");
} // endif

if ( inTrack.numberOfPoints() % pointsPerBank )
{
   throw logic_error ("SST-5136-PFB-PCI: track length can only by n*8 points");
} // endif

if ( (inTrack.startingPoint() & 0xFFFF)
   + (inTrack.numberOfPoints() / pointsPerBank) > maxBytesPerStation )
{
   throw range_error ("SST-5136-PFB-PCI: starting point parameter lower value is too high. SST5136PFBPCI supports up to 244 bytes per slave only.");
} // endif

if ( ((inTrack.startingPoint() >> 16 ) & 0xFFFF) > maxStations )
{
   throw range_error ("SST-5136-PFB-PCI: starting point parameter upper value is too high. SST5136PFBPCI supports up to 126 slaves only.");
} // endif

unsigned long aSlaveId = (inTrack.startingPoint() >> 16) & 0xFFFF;
unsigned short aCrossReff = theMemory -> readShort ( pfbMasStsTab
                                                  + (aSlaveId*2)
                                                  );

if (aCrossReff & 0xFF)
{
   throw range_error ("SST-5136-PFB-PCI: starting point parameter upper value is invalid. Slave is not configured by Profibus.");
}

return * this;
} // end SST5136PFBPCI::addTrack (..)

//-----------------------------------------------------------------------------
void SST5136PFBPCI::executeCommand ( const unsigned char inCmd)
{
BFUNCTRACE_DEVELOP();
  if (inCmd == cmdGO_OFF && state() && state() != cmdON)
     return ;
  while (state() && state() != cmdOFF && !(inCmd == cmdGO_OFF || inCmd == cmdOFF || inCmd == cmdREINIT))
  {
     OSInterface::sleep (10000);
  }
  {
     theMemory -> writeChar (pfbCommand, inCmd);
  }
  do
  {
     OSInterface::sleep (20000);
  }
  while (state() != cmdERROR && state() != cmdOFF && state() != cmdON);
  if (state() == cmdERROR)
  {
     unsigned char errCode = status();
     theMemory -> writeChar (pfbCommand, cmdERR_ACK);
     while ( state() != cmdOFF)
     {
        OSInterface::sleep (20000);
     }
     runtime_error exc (asString("SST-5136-PFB-PCI: Error executing command 0x%X, status 0x%X", inCmd, errCode));
     throw exc;
  }
}
